use std::process::{Command, Stdio};

use serde::{Deserialize, Serialize};
use ropey::Rope;
use tower_lsp::lsp_types::{Diagnostic, Url};

use crate::utils::command::is_command_valid;

use super::{AbstractLinterConfiguration, LinterStatus};

#[derive(Debug, Serialize, Deserialize)]
pub struct IverilogConfiguration {
    pub language_id: String,
    pub linter: IverilogLinter
}

#[derive(Debug, Serialize, Deserialize)]
pub struct IverilogLinter {
    pub name: String,
    pub exe_name: String,
    /// 目前是否启动
    pub enabled: bool,
    pub path: String,
    pub args: Vec<String>,
}

impl IverilogLinter {
    fn new(invoker: &str, args: Vec<String>) -> Self {
        IverilogLinter {
            name: "iverilog".to_string(),
            exe_name: invoker.to_string(),
            enabled: false,
            path: invoker.to_string(),
            args
        }
    }
}

impl AbstractLinterConfiguration for IverilogConfiguration {
    fn new(language_id: &str, invoker: &str, args: Vec<String>) -> Self {
        IverilogConfiguration {
            language_id: language_id.to_string(),
            linter: IverilogLinter::new(invoker, args)
        }
    }

    fn provide_diagnostics(
        &self,
        uri: &Url,
        #[allow(unused)]
        rope: &Rope
    ) -> Vec<Diagnostic> {
        let mut diagnostics = Vec::<Diagnostic>::new();

        let invoke_name = self.get_invoke_name("iverilog");
        let pathbuf = uri.to_file_path().unwrap();
        let path_string = pathbuf.to_str().unwrap();

        let mut process = Command::new(invoke_name)
            .stdin(Stdio::piped())
            .stderr(Stdio::piped())
            .stdout(Stdio::piped())
            .arg("-t null")
            .arg(path_string)
            .spawn();

        
        diagnostics
    }

    fn get_windows_extension(&self) -> &str {
        "exe"
    }

    fn get_linter_path(&self) -> &str {
        &self.linter.path
    }

    fn linter_status(&self) -> LinterStatus {
        match self.language_id.as_str() {
            "verilog" => {
                let invoke_name = self.get_invoke_name("iverilog");
                let available = is_command_valid(&invoke_name);
                LinterStatus {
                    tool_name: "iverilog".to_string(),
                    available,
                    invoke_name
                }
            }

            // 不支持 svlog
            // 不支持 vhdl
            _ => {
                LinterStatus::default()
            }
        }
    }
}
